#! /usr/bin/env python3

import argparse
from pathlib import Path

import rabbitizer
import spimdisasm


def int_any_base(x):
    return int(x, 0)


parser = argparse.ArgumentParser(
    description="Given a rom and start offset, find where the code ends"
)
parser.add_argument("rom", help="path to a .z64 rom")
parser.add_argument("start", help="start offset", type=int_any_base)
parser.add_argument("--end", help="end offset", default=None, type=int_any_base)
parser.add_argument(
    "--vram",
    help="vram address to start disassembly at",
    default="0x80000000",
    type=int_any_base,
)


def run(rom_bytes, start_offset, vram, end_offset=None):
    rom_addr = start_offset
    last_return = rom_addr

    wordList = spimdisasm.common.Utils.bytesToBEWords(rom_bytes[start_offset:])
    for word in wordList:
        insn = rabbitizer.Instruction(word)
        if not insn.isImplemented():
            break

        if insn.isJrRa():
            last_return = rom_addr
        rom_addr += 4
        if end_offset and rom_addr >= end_offset:
            break

    # align to next 0x10 boundary
    end = last_return + 0x10
    end -= end % 0x10
    return end


def main():
    args = parser.parse_args()

    rom_bytes = Path(args.rom).read_bytes()

    start = args.start
    end = args.end
    vram = args.vram

    print(f"0x{run(rom_bytes, start, vram, end):X}")


if __name__ == "__main__":
    main()



#! /usr/bin/env python3

import argparse

import hashlib
import itertools
import struct

import sys
import zlib
from dataclasses import dataclass

from pathlib import Path
from typing import Optional, List

import rabbitizer
import spimdisasm

parser = argparse.ArgumentParser(description="Gives information on N64 roms")
parser.add_argument("rom", help="path to an N64 rom")
parser.add_argument(
    "--header-encoding",
    dest="header_encoding",
    help=(
        "Text encoding the game header is using;"
        " see docs.python.org/3/library/codecs.html#standard-encodings for valid encodings"
    ),
)

country_codes = {
    0x00: "Unknown",
    0x37: "Beta",
    0x41: "Asian (NTSC)",
    0x42: "Brazilian",
    0x43: "Chinese",
    0x44: "German",
    0x45: "North America",
    0x46: "French",
    0x47: "Gateway 64 (NTSC)",
    0x48: "Dutch",
    0x49: "Italian",
    0x4A: "Japanese",
    0x4B: "Korean",
    0x4C: "Gateway 64 (PAL)",
    0x4E: "Canadian",
    0x50: "European (basic spec.)",
    0x53: "Spanish",
    0x55: "Australian",
    0x57: "Scandinavian",
    0x58: "European",
    0x59: "European",
}


@dataclass
class CIC:
    ntsc_name: str
    pal_name: str
    offset: int


crc_to_cic = {
    0x6170A4A1: CIC("6101", "7102", 0x000000),
    0x90BB6CB5: CIC("6102", "7101", 0x000000),
    0x0B050EE0: CIC("6103", "7103", 0x100000),
    0x98BC2C86: CIC("6105", "7105", 0x000000),
    0xACC8580A: CIC("6106", "7106", 0x200000),
}
unknown_cic = CIC("unknown", "unknown", 0x0000000)


@dataclass
class EntryAddressInfo:
    value: int
    rom_hi: int
    rom_lo: int
    ori: bool

    @staticmethod
    def new(
        value: Optional[int], hi: Optional[int], lo: Optional[int], ori: Optional[int]
    ) -> Optional["EntryAddressInfo"]:
        if value is not None and hi is not None and lo is not None:
            return EntryAddressInfo(value, hi, lo, ori == lo)
        return None


@dataclass
class N64EntrypointInfo:
    entry_size: int
    data_size: Optional[int]
    bss_start_address: Optional[EntryAddressInfo]
    bss_size: Optional[EntryAddressInfo]
    bss_end_address: Optional[EntryAddressInfo]
    main_address: Optional[EntryAddressInfo]
    stack_top: Optional[EntryAddressInfo]
    traditional_entrypoint: bool
    ori_entrypoint: bool

    def segment_size(self) -> int:
        if self.data_size is not None:
            return self.entry_size + self.data_size
        return self.entry_size

    def get_bss_size(self) -> Optional[int]:
        if self.bss_size is not None:
            return self.bss_size.value
        if self.bss_start_address is not None and self.bss_end_address is not None:
            return self.bss_end_address.value - self.bss_start_address.value
        return None

    @staticmethod
    def parse_rom_bytes(
        rom_bytes, vram: int, offset: int = 0x1000, size: int = 0x60
    ) -> "N64EntrypointInfo":
        word_list = spimdisasm.common.Utils.bytesToWords(
            rom_bytes, offset, offset + size
        )
        nops_count = 0

        register_values = [0 for _ in range(32)]
        completed_pair = [False for _ in range(32)]
        hi_assignments: List[Optional[int]] = [None for _ in range(32)]
        lo_assignments: List[Optional[int]] = [None for _ in range(32)]
        # We need to track if something was paired using an ori instead of an
        # addiu or similar, because if that's the case we can't emit normal
        # relocations in the generated symbol_addrs file for it.
        ori_assignments: List[Optional[int]] = [None for _ in range(32)]

        register_bss_address: Optional[int] = None
        register_bss_size: Optional[int] = None
        register_main_address: Optional[int] = None

        bss_address: Optional[EntryAddressInfo] = None
        bss_size: Optional[EntryAddressInfo] = None
        bss_end_address: Optional[EntryAddressInfo] = None

        traditional_entrypoint = True
        ori_entrypoint = False
        decrementing_bss_routine = True
        data_size: Optional[int] = None
        func_call_target: Optional[EntryAddressInfo] = None

        size = 0
        i = 0
        while i < len(word_list):
            word = word_list[i]
            current_rom = offset + i * 4
            insn = rabbitizer.Instruction(word, vram)
            if not insn.isValid():
                break

            if insn.isNop():
                nops_count += 1
            elif nops_count >= 3:
                break
            elif insn.canBeHi():
                register_values[insn.rt.value] = insn.getProcessedImmediate() << 16
                completed_pair[insn.rt.value] = False
                hi_assignments[insn.rt.value] = current_rom
            elif insn.canBeLo():
                if insn.isLikelyHandwritten():
                    # Try to skip these instructions:
                    # addi        $t0, $t0, 0x8
                    # addi        $t1, $t1, -0x8
                    pass
                elif insn.modifiesRt() and not completed_pair[insn.rt.value]:
                    register_values[insn.rt.value] = (
                        register_values[insn.rs.value] + insn.getProcessedImmediate()
                    )
                    completed_pair[insn.rt.value] = True
                    lo_assignments[insn.rt.value] = current_rom
                    if insn.isUnsigned():
                        ori_assignments[insn.rt.value] = current_rom
                        ori_entrypoint = True
                elif insn.doesStore():
                    if insn.rt == rabbitizer.RegGprO32.zero:
                        # Try to detect the zero-ing bss algorithm
                        # sw          $zero, 0x0($t0)
                        register_bss_address = insn.rs.value
            elif insn.isBranch():
                if insn.uniqueId == rabbitizer.InstrId.cpu_beq:
                    traditional_entrypoint = False
                    decrementing_bss_routine = False
                elif insn.uniqueId == rabbitizer.InstrId.cpu_bnez:
                    # Traditional entrypoints set the bss size into a register
                    # and loop through it by decrementing it, with a pattern
                    # like the following:
                    #
                    # lui         $t1, %hi(BSS_SIZE)
                    # addiu       $t1, $t1, %lo(BSS_SIZE)
                    # ...
                    # addi        $t1, $t1, -0x8
                    # ...
                    # bnez        $t1, label
                    register_bss_size = insn.rs.value

            elif insn.isJumptableJump() or insn.isReturn():
                # lui         $t2, 0x8000
                # addiu       $t2, $t2, 0x494
                # ...
                # jr          $t2
                register_main_address = insn.rs.value

            elif insn.uniqueId == rabbitizer.InstrId.cpu_sltu:
                # Some non-traditional entrypoints clear bss by loading the
                # bss start and looping though it until reaches the address
                # of the bss end instead of looping by using the bss size
                # explicitly.
                #
                # .clear_bss:
                # addiu      $t0, $t0, 0x4
                # sltu       $at, $t0, $t1
                # bnez       $at, .clear_bss
                #  sw        $zero, -0x4($t0)
                if bss_address is None and bss_size is None:
                    bss_address = EntryAddressInfo.new(
                        register_values[insn.rs.value],
                        hi_assignments[insn.rs.value],
                        lo_assignments[insn.rs.value],
                        ori_assignments[insn.rs.value],
                    )
                    bss_end_address = EntryAddressInfo.new(
                        register_values[insn.rt.value],
                        hi_assignments[insn.rt.value],
                        lo_assignments[insn.rt.value],
                        ori_assignments[insn.rt.value],
                    )

            elif insn.isFunctionCall():
                # Some games don't follow the usual pattern for entrypoints.
                # Those usually use `jal` instead of `jr` to jump out of the
                # entrypoint to actual code.
                traditional_entrypoint = False
                func_call_target = EntryAddressInfo(
                    insn.getInstrIndexAsVram(), current_rom, current_rom, False
                )

            elif insn.uniqueId == rabbitizer.InstrId.cpu_break:
                traditional_entrypoint = False
                size += 4
                vram += 4
                i += 1
                break

            # Traditional entrypoints don't touch the $gp register.
            if insn.modifiesRd() and insn.rd == rabbitizer.RegGprO32.gp:
                traditional_entrypoint = False
            if insn.modifiesRs() and insn.rs == rabbitizer.RegGprO32.gp:
                traditional_entrypoint = False
            if insn.modifiesRt() and insn.rt == rabbitizer.RegGprO32.gp:
                traditional_entrypoint = False

            # print(f"{word:08X}", insn)
            size += 4
            vram += 4
            i += 1

        # for i, val in enumerate(register_values):
        #     if val != 0:
        #         print(i, f"{val:08X}")

        if decrementing_bss_routine:
            if register_bss_address is not None:
                bss_address = EntryAddressInfo.new(
                    register_values[register_bss_address],
                    hi_assignments[register_bss_address],
                    lo_assignments[register_bss_address],
                    ori_assignments[register_bss_address],
                )
            if register_bss_size is not None:
                bss_size = EntryAddressInfo.new(
                    register_values[register_bss_size],
                    hi_assignments[register_bss_size],
                    lo_assignments[register_bss_size],
                    ori_assignments[register_bss_size],
                )

        if register_main_address is not None:
            main_address = EntryAddressInfo.new(
                register_values[register_main_address],
                hi_assignments[register_main_address],
                lo_assignments[register_main_address],
                ori_assignments[register_main_address],
            )
        else:
            main_address = None

        stack_top = EntryAddressInfo.new(
            register_values[rabbitizer.RegGprO32.sp.value],
            hi_assignments[rabbitizer.RegGprO32.sp.value],
            lo_assignments[rabbitizer.RegGprO32.sp.value],
            ori_assignments[rabbitizer.RegGprO32.sp.value],
        )

        if not traditional_entrypoint:
            if func_call_target is not None:
                main_address = func_call_target
                if func_call_target.value > vram:
                    # Some weird-entrypoint games have non-code between the
                    # entrypoint and the actual user code.
                    # We try to find where actual code may begin, and tag
                    # everything in between as "entrypoint data".

                    code_start = find_code_after_data(rom_bytes, offset + i * 4, vram)
                    if code_start is not None and code_start > offset + size:
                        data_size = code_start - (offset + size)

        return N64EntrypointInfo(
            size,
            data_size,
            bss_address,
            bss_size,
            bss_end_address,
            main_address,
            stack_top,
            traditional_entrypoint,
            ori_entrypoint,
        )


def find_code_after_data(
    rom_bytes: bytes, offset: int, vram: int, threshold: int = 0x18000
) -> Optional[int]:
    code_offset: Optional[int] = None

    # We loop through every word until we find a valid `jr $ra` instruction and
    # hope for it to be part of valid code.
    # Once we find it, we loop back until we find anything that is invalid
    # again to try to find the start of this function.

    jr_ra_found = False
    while offset < len(rom_bytes) // 4 and offset < threshold:
        word = spimdisasm.common.Utils.bytesToWords(rom_bytes, offset, offset + 4)[0]
        insn = rabbitizer.Instruction(word, vram)

        if insn.isValid() and insn.isReturn():
            # Check the instruction on the delay slot of the `jr $ra` is valid too.
            next_word = spimdisasm.common.Utils.bytesToWords(
                rom_bytes, offset + 4, offset + 4 + 4
            )[0]
            if rabbitizer.Instruction(next_word, vram + 4).isValid():
                jr_ra_found = True
                break

        vram += 4
        offset += 4

    if jr_ra_found:
        code_offset = offset

        vram -= 4
        offset -= 4

        while offset >= 0:
            word = spimdisasm.common.Utils.bytesToWords(rom_bytes, offset, offset + 4)[
                0
            ]
            insn = rabbitizer.Instruction(word, vram)

            if not insn.isValid():
                # Garbage instructions, stop
                break

            if not insn.isNop():
                # Ignore `nop`s as the code start since they may be file padding.
                code_offset = offset

            vram -= 4
            offset -= 4
    return code_offset


@dataclass
class N64Rom:
    name: str
    header_encoding: str
    country_code: int
    libultra_version: str
    checksum: str
    cic: CIC
    entry_point: int
    size: int
    compiler: str
    sha1: str
    entrypoint_info: N64EntrypointInfo

    def get_country_name(self) -> str:
        return country_codes[self.country_code]


def swap_bytes(data):
    return bytes(
        itertools.chain.from_iterable(
            struct.pack(">H", x) for (x,) in struct.iter_unpack("<H", data)
        )
    )


def read_rom(rom_path: Path):
    rom_bytes = rom_path.read_bytes()

    if rom_path.suffix.lower() == ".n64":
        print("Warning: Input file has .n64 suffix, byte-swapping!")
        rom_bytes = swap_bytes(rom_bytes)
        as_z64 = rom_path.with_suffix(".z64")
        if not as_z64.exists():
            print(f"Writing down {as_z64}")
            as_z64.write_bytes(rom_bytes)
    return rom_bytes


def get_cic(rom_bytes: bytes):
    ipl3_crc = zlib.crc32(rom_bytes[0x40:0x1000])

    return crc_to_cic.get(ipl3_crc, unknown_cic)


def get_entry_point(program_counter: int, cic: CIC):
    return program_counter - cic.offset


def guess_header_encoding(rom_bytes: bytes):
    header = rom_bytes[0x20:0x34]
    encodings = ["ASCII", "shift_jis", "euc-jp"]
    for encoding in encodings:
        try:
            header.decode(encoding)
            return encoding
        except UnicodeDecodeError:
            # we guessed wrong...
            pass

    sys.exit("Unknown header encoding, please raise an Issue with us")


def get_info(
    rom_path: Path, rom_bytes: Optional[bytes] = None, header_encoding=None
) -> N64Rom:
    if rom_bytes is None:
        rom_bytes = read_rom(rom_path)

    if header_encoding is None:
        header_encoding = guess_header_encoding(rom_bytes)

    return get_info_bytes(rom_bytes, header_encoding)


def get_info_bytes(rom_bytes: bytes, header_encoding: str) -> N64Rom:
    (program_counter,) = struct.unpack(">I", rom_bytes[0x8:0xC])
    libultra_version = chr(rom_bytes[0xF])
    checksum = rom_bytes[0x10:0x18].hex().upper()

    try:
        name = rom_bytes[0x20:0x34].decode(header_encoding).rstrip(" \0") or "empty"
    except UnicodeError:
        sys.exit(
            "splat could not decode the game name;"
            " try using a different encoding by passing the --header-encoding argument"
            " (see docs.python.org/3/library/codecs.html#standard-encodings for valid encodings)"
        )

    country_code = rom_bytes[0x3E]

    cic = get_cic(rom_bytes)
    entry_point = get_entry_point(program_counter, cic)

    compiler = get_compiler_info(rom_bytes, entry_point, print_result=False)

    sha1 = hashlib.sha1(rom_bytes).hexdigest()

    entrypoint_info = N64EntrypointInfo.parse_rom_bytes(
        rom_bytes, entry_point, size=0x100
    )

    return N64Rom(
        name,
        header_encoding,
        country_code,
        libultra_version,
        checksum,
        cic,
        entry_point,
        len(rom_bytes),
        compiler,
        sha1,
        entrypoint_info,
    )


def get_compiler_info(rom_bytes, entry_point, print_result=True):
    jumps = 0
    branches = 0

    word_list = spimdisasm.common.Utils.bytesToWords(rom_bytes[0x1000:])
    for word in word_list:
        insn = rabbitizer.Instruction(word)
        if not insn.isImplemented():
            break

        if insn.uniqueId == rabbitizer.InstrId.cpu_j:
            jumps += 1
        elif insn.uniqueId == rabbitizer.InstrId.cpu_b:
            branches += 1

    compiler = "IDO" if branches > jumps else "KMC"
    if print_result:
        print(
            f"{branches} branches and {jumps} jumps detected in the first code segment."
            f" Compiler is most likely {compiler}"
        )
    return compiler


def main():
    rabbitizer.config.pseudos_pseudoB = True

    args = parser.parse_args()
    rom_bytes = read_rom(Path(args.rom))
    rom = get_info(Path(args.rom), rom_bytes, args.header_encoding)

    print("Image name: " + rom.name)
    print("Country code: " + chr(rom.country_code) + " - " + rom.get_country_name())
    print("Libultra version: " + rom.libultra_version)
    print("Checksum: " + rom.checksum)
    print("CIC: " + rom.cic.ntsc_name + " / " + rom.cic.pal_name)
    print("RAM entry point: " + hex(rom.entry_point))
    print("Header encoding: " + rom.header_encoding)
    print("")

    get_compiler_info(rom_bytes, rom.entry_point)


if __name__ == "__main__":
    main()



from ...util import options

from ..common.header import CommonSegHeader


class N64SegHeader(CommonSegHeader):
    def parse_header(self, rom_bytes):
        encoding = options.opts.header_encoding

        header_lines = []
        header_lines.append(".section .data\n")
        header_lines.append(
            self.get_line("word", rom_bytes[0x00:0x04], "PI BSB Domain 1 register")
        )
        header_lines.append(
            self.get_line("word", rom_bytes[0x04:0x08], "Clockrate setting")
        )
        header_lines.append(
            self.get_line("word", rom_bytes[0x08:0x0C], "Entrypoint address")
        )
        header_lines.append(self.get_line("word", rom_bytes[0x0C:0x10], "Revision"))
        header_lines.append(self.get_line("word", rom_bytes[0x10:0x14], "Checksum 1"))
        header_lines.append(self.get_line("word", rom_bytes[0x14:0x18], "Checksum 2"))
        header_lines.append(self.get_line("word", rom_bytes[0x18:0x1C], "Unknown 1"))
        header_lines.append(self.get_line("word", rom_bytes[0x1C:0x20], "Unknown 2"))

        if encoding != "word":
            header_lines.append(
                '.ascii "'
                + rom_bytes[0x20:0x34].decode(encoding).strip().ljust(20)
                + '" /* Internal name */'
            )
        else:
            for i in range(0x20, 0x34, 4):
                header_lines.append(
                    self.get_line("word", rom_bytes[i : i + 4], "Internal name")
                )

        header_lines.append(self.get_line("word", rom_bytes[0x34:0x38], "Unknown 3"))
        header_lines.append(self.get_line("word", rom_bytes[0x38:0x3C], "Cartridge"))
        header_lines.append(
            self.get_line("ascii", rom_bytes[0x3C:0x3E], "Cartridge ID")
        )
        header_lines.append(
            self.get_line("ascii", rom_bytes[0x3E:0x3F], "Country code")
        )
        header_lines.append(self.get_line("byte", rom_bytes[0x3F:0x40], "Version"))
        header_lines.append("")

        return header_lines



from typing import Optional

from ...util import options, symbols, log

from .data import CommonSegData

from ...disassembler.disassembler_section import DisassemblerSection, make_bss_section

# If `options.opts.ld_bss_is_noload` is False, then this segment behaves like a `CommonSegData`


class CommonSegBss(CommonSegData):
    def get_linker_section(self) -> str:
        return ".bss"

    def get_section_flags(self) -> Optional[str]:
        return "wa"

    @staticmethod
    def is_data() -> bool:
        if not options.opts.ld_bss_is_noload:
            return True
        return False

    @staticmethod
    def is_noload() -> bool:
        if not options.opts.ld_bss_is_noload:
            return False
        return True

    def configure_disassembler_section(
        self, disassembler_section: DisassemblerSection
    ) -> None:
        "Allows to configure the section before running the analysis on it"

        pass

    def disassemble_data(self, rom_bytes: bytes):
        if not options.opts.ld_bss_is_noload:
            super().disassemble_data(rom_bytes)
            return

        if self.is_auto_segment:
            return

        if not isinstance(self.rom_start, int):
            log.error(
                f"Segment '{self.name}' (type '{self.type}') requires a rom_start. Got '{self.rom_start}'"
            )

        # Supposedly logic error, not user error
        assert isinstance(self.rom_end, int), f"{self.name} {self.rom_end}"

        # Supposedly logic error, not user error
        segment_rom_start = self.get_most_parent().rom_start
        assert isinstance(segment_rom_start, int), f"{self.name} {segment_rom_start}"

        if not isinstance(self.vram_start, int):
            log.error(
                f"Segment '{self.name}' (type '{self.type}') requires a vram address. Got '{self.vram_start}'"
            )

        next_subsegment = self.parent.get_next_subsegment_for_ram(
            self.vram_start, self.index_within_group
        )
        if next_subsegment is None:
            bss_end = self.get_most_parent().vram_end
        else:
            bss_end = next_subsegment.vram_start
        assert isinstance(bss_end, int), f"{self.name} {bss_end}"

        self.spim_section = make_bss_section(
            self.rom_start,
            self.rom_end,
            self.vram_start,
            bss_end,
            self.name,
            segment_rom_start,
            self.get_exclusive_ram_id(),
        )

        assert self.spim_section is not None

        self.configure_disassembler_section(self.spim_section)

        self.spim_section.analyze()
        self.spim_section.set_comment_offset(self.rom_start)

        for spim_sym in self.spim_section.get_section().symbolList:
            symbols.create_symbol_from_spim_symbol(
                self.get_most_parent(), spim_sym.contextSym, force_in_segment=True
            )

    def should_scan(self) -> bool:
        if not options.opts.ld_bss_is_noload:
            return super().should_scan()
        return options.opts.is_mode_active("code") and self.vram_start is not None



from pathlib import Path
import re
from typing import Optional, TextIO

from ...util import log, options

from .segment import CommonSegment


class CommonSegTextbin(CommonSegment):
    def __init__(
        self,
        rom_start: Optional[int],
        rom_end: Optional[int],
        type: str,
        name: str,
        vram_start: Optional[int],
        args: list,
        yaml,
    ):
        super().__init__(
            rom_start,
            rom_end,
            type,
            name,
            vram_start,
            args=args,
            yaml=yaml,
        )
        self.use_src_path: bool = isinstance(yaml, dict) and yaml.get(
            "use_src_path", False
        )

    @staticmethod
    def is_text() -> bool:
        return True

    def get_linker_section(self) -> str:
        return ".text"

    def get_section_flags(self) -> Optional[str]:
        return "ax"

    def out_path(self) -> Optional[Path]:
        if self.use_src_path:
            return options.opts.src_path / self.dir / f"{self.name}.s"

        return options.opts.data_path / self.dir / f"{self.name}.s"

    def bin_path(self) -> Path:
        typ = self.type
        if typ.startswith("."):
            typ = typ[1:]

        return options.opts.asset_path / self.dir / f"{self.name}.{typ}.bin"

    def write_bin(self, rom_bytes):
        binpath = self.bin_path()
        binpath.parent.mkdir(parents=True, exist_ok=True)

        assert isinstance(self.rom_start, int)
        assert isinstance(self.rom_end, int)

        binpath.write_bytes(rom_bytes[self.rom_start : self.rom_end])

        self.log(f"Wrote {self.name} to {binpath}")

    def write_asm_contents(self, rom_bytes, f: TextIO):
        binpath = self.bin_path()
        asm_label = options.opts.asm_function_macro
        if not self.is_text():
            asm_label = options.opts.asm_data_macro

        assert isinstance(self.rom_start, int)
        assert isinstance(self.rom_end, int)

        f.write(f"{self.get_section_asm_line()}\n\n")

        sym_name = None
        sym_name_end = None
        sym_size_matches = None

        # Check if there's a symbol at this address
        vram = self.rom_to_ram(self.rom_start)
        if vram is not None:
            sym = self.get_symbol(vram, in_segment=True)
            if sym is not None:
                sym.defined = True
                sym_name = sym.name
                sym_name_end = sym.given_name_end
                if (
                    sym.given_size is None
                    or sym.given_size == self.rom_end - self.rom_start
                ):
                    sym_size_matches = self.rom_end - self.rom_start

        if sym_name is None:
            # Normalize stuff like slashes and such.
            n = regex_sym_name_normalizer.sub("_", self.name)
            if self.is_text():
                suffix = "textbin"
            elif self.is_data():
                suffix = "databin"
            elif self.is_rodata():
                suffix = "rodatabin"
            else:
                suffix = "incbin"
            sym_name = f"__{n}_{suffix}"

        if options.opts.asm_nonmatching_label_macro != "":
            siz = f", 0x{sym_size_matches:X}" if sym_size_matches is not None else ""
            f.write(f"{options.opts.asm_nonmatching_label_macro} {sym_name}{siz}\n\n")

        f.write(f"{asm_label} {sym_name}\n")
        if asm_label == ".globl":
            if self.is_text():
                f.write(f".ent {sym_name}\n")
            f.write(f"{sym_name}:\n")

        f.write(f'.incbin "{binpath.as_posix()}"\n')

        if options.opts.asm_emit_size_directive:
            f.write(f".size {sym_name}, . - {sym_name}\n")

        if self.is_text() and options.opts.asm_end_label != "":
            f.write(f"{options.opts.asm_end_label} {sym_name}\n")
        elif options.opts.asm_data_end_label != "":
            f.write(f"{options.opts.asm_data_end_label} {sym_name}\n")

        if sym_name_end is not None and sym_size_matches is not None:
            f.write(f"{asm_label} {sym_name_end}\n")
            if asm_label == ".globl":
                f.write(f"{sym_name_end}:\n")

            if self.is_text() and options.opts.asm_end_label != "":
                f.write(f"{options.opts.asm_end_label} {sym_name_end}\n")
            elif options.opts.asm_data_end_label != "":
                f.write(f"{options.opts.asm_data_end_label} {sym_name_end}\n")

    def split(self, rom_bytes):
        if self.rom_end is None:
            log.error(
                f"segment {self.name} needs to know where it ends; add a position marker [0xDEADBEEF] after it"
            )

        self.write_bin(rom_bytes)

        s_path = self.out_path()
        assert s_path is not None

        if s_path.exists():
            return

        s_path.parent.mkdir(parents=True, exist_ok=True)

        with s_path.open("w", encoding="utf-8") as f:
            f.write('.include "macro.inc"\n\n')
            preamble = options.opts.generated_s_preamble
            if preamble:
                f.write(preamble + "\n")

            self.write_asm_contents(rom_bytes, f)

            for sect in self.section_order:
                if sect == self.get_linker_section_linksection():
                    continue

                sibling = self.siblings.get(sect)
                if sibling is None:
                    continue

                # We check against CommonSegTextbin instead of the specific type because the other incbins inherit from this class
                if isinstance(sibling, CommonSegTextbin):
                    f.write("\n")
                    sibling.write_asm_contents(rom_bytes, f)

    def should_scan(self) -> bool:
        return self.rom_start is not None and self.rom_end is not None

    def should_split(self) -> bool:
        return (
            self.extract and self.should_scan()
        )  # only split if the segment was scanned first


regex_sym_name_normalizer = re.compile(r"[^0-9a-zA-Z_]")



from abc import ABC, abstractmethod
from typing import Set


class Disassembler(ABC):
    @abstractmethod
    def configure(self):
        raise NotImplementedError("configure")

    @abstractmethod
    def check_version(self, skip_version_check: bool, splat_version: str):
        raise NotImplementedError("check_version")

    @abstractmethod
    def known_types(self) -> Set[str]:
        raise NotImplementedError("known_types")



"""
This module is used to load splat configuration from a YAML file.

A config dict can be loaded using `load`.

    config = conf.load("path/to/splat.yaml")
"""

from typing import Any, Dict, List, Optional
from pathlib import Path

# This unused import makes the yaml library faster. don't remove
import pylibyaml  # noqa: F401
import yaml

from . import options, vram_classes


def _merge_configs(main_config, additional_config, additional_config_path):
    # Merge rules are simple
    # For each key in the dictionary
    # - If list then append to list
    # - If a dictionary then repeat merge on sub dictionary entries
    # - Else assume string or number and replace entry

    for curkey in additional_config:
        if curkey not in main_config:
            main_config[curkey] = additional_config[curkey]
        elif type(main_config[curkey]) is not type(additional_config[curkey]):
            raise TypeError(
                f"Could not merge {str(additional_config_path)}: type for key '{curkey}' in configs does not match"
            )
        else:
            # keys exist and match, see if a list to append
            if type(main_config[curkey]) is list:
                main_config[curkey] += additional_config[curkey]
            elif type(main_config[curkey]) is dict:
                # need to merge sub areas
                main_config[curkey] = _merge_configs(
                    main_config[curkey],
                    additional_config[curkey],
                    additional_config_path,
                )
            else:
                # not a list or dictionary, must be a number or string, overwrite
                main_config[curkey] = additional_config[curkey]

    return main_config


def load(
    config_path: List[Path],
    modes: Optional[List[str]] = None,
    verbose: bool = False,
    disassemble_all: bool = False,
    make_full_disasm_for_code=False,
) -> Dict[str, Any]:
    """
    Returns a `dict` with resolved splat config.

    Multiple configuration files can be passed in ``config_path`` with each subsequent file merged into the previous.

    `modes` specifies which modes are active (all, code, img, gfx, vtx, etc.). The default is all.

    `verbose` may be used to determine whether or not to display additional output.

    `disassemble_all` determines whether functions which are already compiled will be disassembled. This is OR-ed with
    the `disassemble_all` key in a config file, if present.

    After all files are merged, static validation is done on the configuration.

    The returned `dict` represents the merged and validated YAML config.

    As a side effect, the global `splat.util.options.opts` is set.

    Config with invalid options may raise an error.
    """

    config: Dict[str, Any] = {}
    for entry in config_path:
        entry_text = entry.read_text(encoding="utf-8")
        additional_config = yaml.load(entry_text, Loader=yaml.SafeLoader)
        config = _merge_configs(config, additional_config, entry)

    vram_classes.initialize(config.get("vram_classes"))

    options.initialize(
        config,
        config_path,
        modes,
        verbose,
        disassemble_all,
        make_full_disasm_for_code,
    )

    return config



from ..common.hasm import CommonSegHasm


class N64SegIpl3(CommonSegHasm):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.special_vram_segment = True



from typing import Optional

from ...util import log, options

from .textbin import CommonSegTextbin


class CommonSegRodatabin(CommonSegTextbin):
    @staticmethod
    def is_text() -> bool:
        return False

    @staticmethod
    def is_rodata() -> bool:
        return True

    def get_linker_section(self) -> str:
        return ".rodata"

    def get_section_flags(self) -> Optional[str]:
        return "a"

    def split(self, rom_bytes):
        if self.rom_end is None:
            log.error(
                f"segment {self.name} needs to know where it ends; add a position marker [0xDEADBEEF] after it"
            )

        self.write_bin(rom_bytes)

        if self.sibling is None:
            # textbin will write the incbin instead

            s_path = self.out_path()
            assert s_path is not None
            s_path.parent.mkdir(parents=True, exist_ok=True)

            with s_path.open("w", encoding="utf-8") as f:
                f.write('.include "macro.inc"\n\n')
                preamble = options.opts.generated_s_preamble
                if preamble:
                    f.write(preamble + "\n")

                self.write_asm_contents(rom_bytes, f)



from pathlib import Path
from typing import Optional, List

from ...util import log, options

from ..linker_entry import LinkerEntry, LinkerWriter
from .segment import CommonSegment

from ..segment import Segment, parse_segment_vram


class LinkerEntryLib(LinkerEntry):
    def __init__(
        self,
        segment: Segment,
        src_paths: List[Path],
        object_path: Path,
        section_order: str,
        section_link: str,
        noload: bool,
    ):
        super().__init__(
            segment, src_paths, object_path, section_order, section_link, noload
        )
        self.object_path = object_path

    def emit_entry(self, linker_writer: LinkerWriter):
        self.emit_path(linker_writer)


class CommonSegLib(CommonSegment):
    def __init__(
        self,
        rom_start: Optional[int],
        rom_end: Optional[int],
        type: str,
        name: str,
        vram_start: Optional[int],
        args: list,
        yaml,
    ):
        super().__init__(
            rom_start,
            rom_end,
            type,
            name,
            vram_start,
            args=args,
            yaml=yaml,
        )

        vram = parse_segment_vram(self.yaml)
        if vram is not None:
            self.vram_start = vram

        if isinstance(yaml, dict):
            self.object = yaml.get("object", None)
            self.section = yaml.get("section", ".text")

            if self.object is None:
                log.error(f"Error: {self.name} is missing object file")
        else:
            if len(args) < 1:
                log.error(f"Error: {self.name} is missing object file")

            if len(args) > 1:
                self.object, self.section = args[0], args[1]
            else:
                self.object, self.section = args[0], ".text"

        self.extract = False

    def get_linker_section(self) -> str:
        return self.section

    def get_linker_entries(self) -> List[LinkerEntry]:
        path = options.opts.lib_path / self.name

        object_path = Path(f"{path}.a:{self.object}.o")

        return [
            LinkerEntryLib(
                self,
                [path],
                object_path,
                self.get_linker_section_order(),
                self.get_linker_section_linksection(),
                self.is_noload(),
            )
        ]



from typing import Optional

from .data import CommonSegData
from ...util import log

from ...disassembler.disassembler_section import (
    DisassemblerSection,
    make_gcc_except_table_section,
)


class CommonSegGcc_except_table(CommonSegData):
    """Segment containing the GCC Except Table (aka ehtable), used for implementing C++ exceptions"""

    def get_linker_section(self) -> str:
        return ".gcc_except_table"

    def get_section_flags(self) -> Optional[str]:
        return "aw"

    def configure_disassembler_section(
        self, disassembler_section: DisassemblerSection
    ) -> None:
        "Allows to configure the section before running the analysis on it"

        super().configure_disassembler_section(disassembler_section)

        section = disassembler_section.get_section()

        section.enableStringGuessing = False

    def disassemble_data(self, rom_bytes):
        if self.is_auto_segment:
            return

        if not isinstance(self.rom_start, int):
            log.error(
                f"Segment '{self.name}' (type '{self.type}') requires a rom_start. Got '{self.rom_start}'"
            )

        # Supposedly logic error, not user error
        assert isinstance(self.rom_end, int), self.rom_end

        # Supposedly logic error, not user error
        segment_rom_start = self.get_most_parent().rom_start
        assert isinstance(segment_rom_start, int), segment_rom_start

        if not isinstance(self.vram_start, int):
            log.error(
                f"Segment '{self.name}' (type '{self.type}') requires a vram address. Got '{self.vram_start}'"
            )

        self.spim_section = make_gcc_except_table_section(
            self.rom_start,
            self.rom_end,
            self.vram_start,
            self.name,
            rom_bytes,
            segment_rom_start,
            self.get_exclusive_ram_id(),
        )

        assert self.spim_section is not None

        self.configure_disassembler_section(self.spim_section)

        self.spim_section.analyze()
        self.spim_section.set_comment_offset(self.rom_start)



import sys
from typing import NoReturn, Optional
from pathlib import Path

from colorama import Fore, init, Style

init(autoreset=True)

newline = True

Status = Optional[str]


def write(*args, status=None, **kwargs):
    global newline

    if not newline:
        print("")
        newline = True

    print(
        status_to_ansi(status) + str(args[0]),
        *args[1:],
        **kwargs,
        file=output_file(status),
    )


def error(*args, **kwargs) -> NoReturn:
    write(*args, **kwargs, status="error")
    sys.exit(2)


# The line_num is expected to be zero-indexed
def parsing_error_preamble(path: Path, line_num: int, line: str):
    write("")
    write(f"error reading {path}, line {line_num + 1}:", status="error")
    write(f"\t{line}")


def status_to_ansi(status: Status):
    if status == "ok":
        return Fore.GREEN
    elif status == "warn":
        return Fore.YELLOW + Style.BRIGHT
    elif status == "error":
        return Fore.RED + Style.BRIGHT
    elif status == "skip":
        return Style.DIM
    else:
        return ""


def output_file(status: Status):
    if status == "warn" or status == "error":
        return sys.stderr
    return sys.stdout



import pickle
from typing import Any, Dict

from . import options, log
from ..segtypes.common.segment import Segment


class Cache:
    def __init__(self, config: Dict[str, Any], use_cache: bool, verbose: bool):
        self.use_cache: bool = use_cache
        self.cache: Dict[str, Any] = {}

        # Load cache
        if use_cache and options.opts.cache_path.exists():
            try:
                with options.opts.cache_path.open("rb") as f3:
                    self.cache = pickle.load(f3)

                if verbose:
                    log.write(f"Loaded cache ({len(self.cache.keys())} items)")
            except Exception:
                log.write(
                    "Not able to load cache file. Discarding old cache", status="warn"
                )

        # invalidate entire cache if options change
        if use_cache and self.cache.get("__options__") != config.get("options"):
            if verbose:
                log.write("Options changed, invalidating cache")

            self.cache = {
                "__options__": config.get("options"),
            }

    def save(self, verbose: bool):
        if self.cache != {} and self.use_cache:
            if verbose:
                log.write("Writing cache")
            options.opts.cache_path.parent.mkdir(parents=True, exist_ok=True)
            with options.opts.cache_path.open("wb") as f4:
                pickle.dump(self.cache, f4)

    def check_cache_hit(self, segment: Segment, update_on_miss: bool) -> bool:
        if self.use_cache:
            cached = segment.cache()
            segment_id = segment.unique_id()

            if cached == self.cache.get(segment_id):
                # Cache hit
                return True

            # Cache miss
            if update_on_miss:
                self.cache[segment_id] = cached

        return False



from . import disassembler
from typing import Set


class NullDisassembler(disassembler.Disassembler):
    def configure(self):
        pass

    def check_version(self, skip_version_check: bool, splat_version: str):
        pass

    def known_types(self) -> Set[str]:
        return set()



from colorama import Fore, Style
from typing import Dict, Optional

from . import log


def fmt_size(size: int) -> str:
    if size > 1000000:
        return str(size // 1000000) + " MB"
    elif size > 1000:
        return str(size // 1000) + " KB"
    else:
        return str(size) + " B"


class Statistics:
    def __init__(self):
        self.seg_sizes: Dict[str, int] = {}
        self.seg_split: Dict[str, int] = {}
        self.seg_cached: Dict[str, int] = {}

    def add_size(self, typ: str, size: Optional[int]):
        if typ not in self.seg_sizes:
            self.seg_sizes[typ] = 0
        self.seg_sizes[typ] += 0 if size is None else size

    def count_split(self, typ: str, count: int = 1):
        if typ not in self.seg_split:
            self.seg_split[typ] = 0
        self.seg_split[typ] += count

    def count_cached(self, typ: str, count: int = 1):
        if typ not in self.seg_cached:
            self.seg_cached[typ] = 0
        self.seg_cached[typ] += count

    def print_statistics(self, total_size: int):
        unk_size = self.seg_sizes.get("unk", 0)
        rest_size = 0

        for typ, size in self.seg_sizes.items():
            if typ != "unk":
                rest_size += size

        known_ratio = rest_size / total_size
        unk_ratio = unk_size / total_size

        log.write(
            f"Split {fmt_size(rest_size)} ({known_ratio:.2%}) in defined segments"
        )
        for typ, size in self.seg_sizes.items():
            if typ != "unk":
                tmp_ratio = size / total_size
                log.write(
                    f"{typ:>20}: {fmt_size(size):>8} ({tmp_ratio:.2%}) {Fore.GREEN}{self.seg_split.get(typ, 0)} split{Style.RESET_ALL}, {Style.DIM}{self.seg_cached.get(typ, 0)} cached"
                )
        log.write(
            f"{'unknown':>20}: {fmt_size(unk_size):>8} ({unk_ratio:.2%}) from unknown bin files"
        )



from collections import OrderedDict
from typing import List, Optional, Type, Tuple

from ...util import log, options, utils

from .group import CommonSegGroup
from ..segment import Segment, parse_segment_align


def dotless_type(type: str) -> str:
    return type[1:] if type[0] == "." else type


# code group
class CommonSegCode(CommonSegGroup):
    def __init__(
        self,
        rom_start: Optional[int],
        rom_end: Optional[int],
        type: str,
        name: str,
        vram_start: Optional[int],
        args: list,
        yaml,
    ):
        self.bss_size: int = yaml.get("bss_size", 0) if isinstance(yaml, dict) else 0

        super().__init__(
            rom_start,
            rom_end,
            type,
            name,
            vram_start,
            args=args,
            yaml=yaml,
        )

        self.reported_file_split = False

        self.align = parse_segment_align(yaml)
        if self.align is None:
            self.align = 0x10

    @property
    def needs_symbols(self) -> bool:
        return True

    @property
    def vram_end(self) -> Optional[int]:
        if self.vram_start is not None and self.size is not None:
            return self.vram_start + self.size + self.bss_size
        else:
            return None

    # Generates a placeholder segment for the auto_link_sections option
    def _generate_segment_from_all(
        self,
        rep_type: str,
        replace_class: Type[Segment],
        base_name: str,
        base_seg: Segment,
        rom_start: Optional[int] = None,
        rom_end: Optional[int] = None,
        vram_start: Optional[int] = None,
    ) -> Segment:
        rep: Segment = replace_class(
            rom_start=rom_start,
            rom_end=rom_end,
            type=rep_type,
            name=base_name,
            vram_start=vram_start,
            args=[],
            yaml={},
        )
        rep.extract = False
        rep.given_subalign = self.given_subalign
        rep.exclusive_ram_id = self.get_exclusive_ram_id()
        rep.given_dir = self.given_dir
        rep.given_symbol_name_format = self.symbol_name_format
        rep.given_symbol_name_format_no_rom = self.symbol_name_format_no_rom
        assert rep != base_seg, rep
        rep.sibling = base_seg
        rep.parent = self
        rep.is_generated = True
        rep.is_auto_segment = True
        if rep.special_vram_segment:
            self.special_vram_segment = True
        rep.bss_contains_common = self.bss_contains_common
        return rep

    def _insert_all_auto_sections(
        self,
        ret: List[Segment],
        base_segments: OrderedDict[str, Segment],
        readonly_before: bool,
    ) -> List[Segment]:
        if len(options.opts.auto_link_sections) == 0:
            return ret

        base_segments_list: List[Tuple[str, Segment]] = list(base_segments.items())

        # Determine what will be the min insertion index
        last_inserted_index = len(base_segments_list) - 1
        for i, seg in enumerate(ret):
            if seg.is_text():
                last_inserted_index = i

            elif readonly_before:
                if seg.is_rodata():
                    last_inserted_index = i

        for i, sect in enumerate(options.opts.auto_link_sections):
            for name, seg in base_segments_list:
                link_section = seg.get_linker_section_order()
                if link_section == sect or link_section == "":
                    # Avoid duplicating current section
                    # and advance over files without explicit sections (like linker_offset)
                    last_inserted_index = ret.index(seg)
                    continue

                sibling = seg.siblings.get(sect)
                if sibling is None:
                    replace_class = Segment.get_class_for_type(sect)
                    sibling = self._generate_segment_from_all(
                        sect, replace_class, seg.name, seg
                    )
                    seg.siblings[sect] = sibling
                    last_inserted_index += 1
                    ret.insert(last_inserted_index, sibling)

                last_inserted_index = ret.index(sibling)

            # Advance last_inserted_index for any segment of a type that we have already seen
            while last_inserted_index < len(ret):
                link_section = ret[last_inserted_index].get_linker_section_order()
                if (
                    link_section != ""
                    and link_section not in options.opts.auto_link_sections[: i + 1]
                ):
                    last_inserted_index -= 1
                    if last_inserted_index < 0:
                        last_inserted_index = 0
                    break
                last_inserted_index += 1

        return ret

    def parse_subsegments(self, segment_yaml) -> List[Segment]:
        if "subsegments" not in segment_yaml:
            if not self.parent:
                raise Exception(
                    f"No subsegments provided in top-level code segment {self.name}"
                )
            return []

        base_segments: OrderedDict[str, Segment] = OrderedDict()
        ret: List[Segment] = []
        prev_start: Optional[int] = -1
        prev_vram: Optional[int] = -1

        last_rom_end = None

        # Determine what comes first, either text or rodata/rdata
        readonly_before = False
        text_index = utils.list_index(self.section_order, ".text")
        if text_index is not None:
            rodata_index = utils.list_index(self.section_order, ".rodata")
            if rodata_index is not None:
                readonly_before = rodata_index < text_index
            else:
                rdata_index = utils.list_index(self.section_order, ".rdata")
                if rdata_index is not None:
                    readonly_before = rdata_index < text_index

        for i, subsegment_yaml in enumerate(segment_yaml["subsegments"]):
            # endpos marker
            if isinstance(subsegment_yaml, list) and len(subsegment_yaml) == 1:
                continue

            typ = Segment.parse_segment_type(subsegment_yaml)
            start, is_auto_segment = Segment.parse_segment_start(subsegment_yaml)

            segment_class = Segment.get_class_for_type(typ)

            if start is None:
                # Attempt to infer the start address
                if i == 0:
                    # The start address of this segment is the start address of the group
                    start = self.rom_start
                else:
                    # The start address is the end address of the previous segment
                    start = last_rom_end

            # First, try to get the end address from the next segment's start address
            # Second, try to get the end address from the estimated size of this segment
            # Third, try to get the end address from the next segment with a start address
            end: Optional[int] = None
            if i < len(segment_yaml["subsegments"]) - 1:
                end, end_is_auto_segment = Segment.parse_segment_start(
                    segment_yaml["subsegments"][i + 1]
                )
            if start is not None and end is None:
                est_size = segment_class.estimate_size(subsegment_yaml)
                if est_size is not None:
                    end = start + est_size
            if end is None:
                end = self.get_next_seg_start(i, segment_yaml["subsegments"])

            if start is not None and prev_start is not None and start < prev_start:
                log.error(
                    f"Error: Group segment '{self.name}' contains subsegments which are out of ascending rom order (0x{prev_start:X} followed by 0x{start:X})"
                )

            vram = None
            if start is not None:
                assert isinstance(start, int)
                vram = self.get_most_parent().rom_to_ram(start)

            if segment_class.is_noload() and last_rom_end is not None:
                # Pretend bss's rom address is after the last actual rom segment
                start = last_rom_end
                # and it has a rom size of zero
                end = last_rom_end

            segment: Segment = Segment.from_yaml(
                segment_class, subsegment_yaml, start, end, self, vram
            )
            segment.is_auto_segment = is_auto_segment

            if (
                segment.vram_start is not None
                and prev_vram is not None
                and segment.vram_start < prev_vram
            ):
                log.error(
                    f"Error: Group segment '{self.name}' contains subsegments which are out of ascending vram order (0x{prev_vram:X} followed by 0x{segment.vram_start:X}).\n"
                    + f"Detected when processing file '{segment.name}' of type '{segment.type}'"
                )

            ret.append(segment)

            if segment.is_text():
                base_segments[segment.name] = segment

            if readonly_before:
                if segment.is_rodata() and segment.sibling is None:
                    base_segments[segment.name] = segment

            if segment.special_vram_segment:
                self.special_vram_segment = True

            segment.bss_contains_common = self.bss_contains_common

            prev_start = start
            prev_vram = segment.vram_start
            if end is not None:
                last_rom_end = end

        # We need base_segments to be fully constructed before start assigning siblings to segments,
        # otherwise we may miss segments that are placed before any text segment
        for segment in ret:
            sibling = base_segments.get(segment.name, None)
            if sibling == segment:
                sibling = None
            segment.sibling = sibling

            if sibling is not None:
                # Make siblings reference between them
                segment.siblings[sibling.get_linker_section_linksection()] = sibling
                sibling.siblings[segment.get_linker_section_linksection()] = segment

        ret = self._insert_all_auto_sections(ret, base_segments, readonly_before)
        for i, seg in enumerate(ret):
            seg.index_within_group = i

        return ret

    def scan(self, rom_bytes):
        # Always scan code first
        for sub in self.subsegments:
            if sub.is_text() and sub.should_scan():
                sub.scan(rom_bytes)

        # Scan everyone else
        for sub in self.subsegments:
            if not sub.is_text() and sub.should_scan():
                sub.scan(rom_bytes)



from . import ci as ci
from . import ci4 as ci4
from . import ci8 as ci8
from . import decompressor as decompressor
from . import gfx as gfx
from . import header as header
from . import i1 as i1
from . import i4 as i4
from . import i8 as i8
from . import ia16 as ia16
from . import ia4 as ia4
from . import ia8 as ia8
from . import img as img
from . import ipl3 as ipl3
from . import mio0 as mio0
from . import palette as palette
from . import rgba16 as rgba16
from . import rgba32 as rgba32
from . import rsp as rsp
from . import vtx as vtx
from . import yay0 as yay0
